package adapter

import (
	"context"
	"database/sql"
	"errors"
	"fmt"
	"time"

	"gitea.dwysokinski.me/twhelp/corev3/internal/adapter/internal/bunmodel"
	"gitea.dwysokinski.me/twhelp/corev3/internal/domain"
	"github.com/uptrace/bun"
	"github.com/uptrace/bun/dialect"
)

type ServerBunRepository struct {
	db bun.IDB
}

func NewServerBunRepository(db bun.IDB) *ServerBunRepository {
	return &ServerBunRepository{db: db}
}

func (repo *ServerBunRepository) CreateOrUpdate(ctx context.Context, params ...domain.CreateServerParams) error {
	if len(params) == 0 {
		return nil
	}

	servers := make(bunmodel.Servers, 0, len(params))

	for _, p := range params {
		base := p.Base()
		servers = append(servers, bunmodel.Server{
			Key:         base.Key(),
			URL:         base.URL().String(),
			Open:        base.Open(),
			VersionCode: p.VersionCode(),
			CreatedAt:   time.Now(),
		})
	}

	q := repo.db.NewInsert().
		Model(&servers)

	//nolint:exhaustive
	switch q.Dialect().Name() {
	case dialect.PG:
		q = q.On("CONFLICT ON CONSTRAINT servers_pkey DO UPDATE")
	case dialect.SQLite:
		q = q.On("CONFLICT(key) DO UPDATE")
	default:
		q = q.Err(errors.New("unsupported dialect"))
	}

	if _, err := q.
		Set("url = EXCLUDED.url").
		Set("open = EXCLUDED.open").
		Returning("").
		Exec(ctx); err != nil {
		return fmt.Errorf("something went wrong while inserting servers into the db: %w", err)
	}

	return nil
}

func (repo *ServerBunRepository) List(ctx context.Context, params domain.ListServersParams) (domain.Servers, error) {
	var servers bunmodel.Servers

	if err := repo.baseListQuery(params).Model(&servers).Scan(ctx); err != nil && !errors.Is(err, sql.ErrNoRows) {
		return nil, fmt.Errorf("couldn't select servers from the db: %w", err)
	}

	return servers.ToDomain()
}

func (repo *ServerBunRepository) baseListQuery(params domain.ListServersParams) *bun.SelectQuery {
	return repo.db.NewSelect().Apply(listServersParamsApplier{params: params}.apply)
}

type listServersParamsApplier struct {
	params domain.ListServersParams
}

func (a listServersParamsApplier) apply(q *bun.SelectQuery) *bun.SelectQuery {
	if keys := a.params.Keys(); len(keys) > 0 {
		q = q.Where("server.key IN (?)", bun.In(keys))
	}

	if keyGT := a.params.KeyGT(); keyGT.Valid {
		q = q.Where("server.key > ?", keyGT.Value)
	}

	if open := a.params.Open(); open.Valid {
		q = q.Where("server.open = ?", open.Value)
	}

	if special := a.params.Special(); special.Valid {
		q = q.Where("server.special = ?", special.Value)
	}

	q = q.Limit(a.params.Limit()).Offset(a.params.Offset())

	for _, s := range a.params.Sort() {
		switch s {
		case domain.ServerSortKeyASC:
			q = q.Order("server.key ASC")
		case domain.ServerSortKeyDESC:
			q = q.Order("server.key DESC")
		case domain.ServerSortOpenASC:
			q = q.Order("server.open ASC")
		case domain.ServerSortOpenDESC:
			q = q.Order("server.open DESC")
		default:
			return q.Err(errors.New("unsupported sort value"))
		}
	}

	return q
}
