package domain_test

import (
	"fmt"
	"net/url"
	"testing"

	"gitea.dwysokinski.me/twhelp/corev3/internal/domain"
	"gitea.dwysokinski.me/twhelp/corev3/internal/domain/domaintest"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
)

func TestNewSyncServersCmdPayload(t *testing.T) {
	t.Parallel()

	type args struct {
		versionCode string
		url         *url.URL
	}

	tests := []struct {
		name        string
		args        args
		expectedErr error
	}{
		{
			name: "OK",
			args: args{
				versionCode: "pl",
				url: &url.URL{
					Scheme: "https",
					Host:   "plemiona.pl",
				},
			},
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			payload, err := domain.NewSyncServersCmdPayload(tt.args.versionCode, tt.args.url)
			require.ErrorIs(t, err, tt.expectedErr)
			if tt.expectedErr != nil {
				return
			}
			assert.Equal(t, tt.args.versionCode, payload.VersionCode())
			assert.Equal(t, tt.args.url, payload.URL())
		})
	}
}

func TestNewSyncServersCmdPayloadWithStringURL(t *testing.T) {
	t.Parallel()

	type args struct {
		versionCode string
		url         string
	}

	tests := []struct {
		name        string
		args        args
		expectedErr error
	}{
		{
			name: "OK",
			args: args{
				versionCode: "pl",
				url:         "https://plemiona.pl",
			},
		},
		{
			name: "ERR: invalid url",
			args: args{
				versionCode: "pl",
				url:         "plemiona.pl",
			},
			expectedErr: domain.InvalidURLError{
				URL: "plemiona.pl",
			},
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			payload, err := domain.NewSyncServersCmdPayloadWithStringURL(tt.args.versionCode, tt.args.url)
			require.ErrorIs(t, err, tt.expectedErr)
			if tt.expectedErr != nil {
				return
			}
			assert.Equal(t, tt.args.versionCode, payload.VersionCode())
			assert.Equal(t, tt.args.url, payload.URL().String())
		})
	}
}

func TestNewCreateServerParams(t *testing.T) {
	t.Parallel()

	validVersion := domaintest.NewVersion(t, domaintest.VersionConfig{})
	validBaseServer := domaintest.NewBaseServer(t, domaintest.BaseServerConfig{})

	type args struct {
		servers     domain.BaseServers
		versionCode string
	}

	type test struct {
		name        string
		args        args
		expectedErr error
	}

	tests := []test{
		{
			name: "OK",
			args: args{
				versionCode: validVersion.Code(),
				servers: domain.BaseServers{
					validBaseServer,
				},
			},
		},
	}

	for _, versionCodeTest := range newVersionCodeValidationTests() {
		tests = append(tests, test{
			name: versionCodeTest.name,
			args: args{
				versionCode: versionCodeTest.code,
				servers: domain.BaseServers{
					validBaseServer,
				},
			},
			expectedErr: domain.ValidationError{
				Err:   versionCodeTest.expectedErr,
				Field: "versionCode",
			},
		})
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			res, err := domain.NewCreateServerParams(tt.args.servers, tt.args.versionCode)
			require.ErrorIs(t, err, tt.expectedErr)
			if tt.expectedErr != nil {
				return
			}
			require.Len(t, res, len(tt.args.servers))
			for i, b := range tt.args.servers {
				assert.Equal(t, b, res[i].Base())
				assert.Equal(t, tt.args.versionCode, res[i].VersionCode())
			}
		})
	}
}

func TestListServersParams_SetSort(t *testing.T) {
	t.Parallel()

	type args struct {
		sort []domain.ServerSort
	}

	tests := []struct {
		name        string
		args        args
		expectedErr error
	}{
		{
			name: "OK",
			args: args{
				sort: []domain.ServerSort{
					domain.ServerSortKeyASC,
				},
			},
		},
		{
			name: "ERR: len(sort) < 1",
			args: args{
				sort: nil,
			},
			expectedErr: domain.ValidationError{
				Field: "sort",
				Err: domain.LenOutOfRangeError{
					Min:     1,
					Max:     2,
					Current: 0,
				},
			},
		},
		{
			name: "ERR: len(sort) > 2",
			args: args{
				sort: []domain.ServerSort{
					domain.ServerSortKeyASC,
					domain.ServerSortKeyDESC,
					domain.ServerSortOpenASC,
				},
			},
			expectedErr: domain.ValidationError{
				Field: "sort",
				Err: domain.LenOutOfRangeError{
					Min:     1,
					Max:     2,
					Current: 3,
				},
			},
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			params := domain.NewListServersParams()

			require.ErrorIs(t, params.SetSort(tt.args.sort), tt.expectedErr)
			if tt.expectedErr != nil {
				return
			}
			assert.Equal(t, tt.args.sort, params.Sort())
		})
	}
}

func TestListServersParams_SetLimit(t *testing.T) {
	t.Parallel()

	type args struct {
		limit int
	}

	tests := []struct {
		name        string
		args        args
		expectedErr error
	}{
		{
			name: "OK",
			args: args{
				limit: domain.ServerListMaxLimit,
			},
		},
		{
			name: "ERR: limit < 1",
			args: args{
				limit: 0,
			},
			expectedErr: domain.ValidationError{
				Field: "limit",
				Err: domain.MinGreaterEqualError{
					Min:     1,
					Current: 0,
				},
			},
		},
		{
			name: fmt.Sprintf("ERR: limit > %d", domain.ServerListMaxLimit),
			args: args{
				limit: domain.ServerListMaxLimit + 1,
			},
			expectedErr: domain.ValidationError{
				Field: "limit",
				Err: domain.MaxLessEqualError{
					Max:     domain.ServerListMaxLimit,
					Current: domain.ServerListMaxLimit + 1,
				},
			},
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			params := domain.NewListServersParams()

			require.ErrorIs(t, params.SetLimit(tt.args.limit), tt.expectedErr)
			if tt.expectedErr != nil {
				return
			}
			assert.Equal(t, tt.args.limit, params.Limit())
		})
	}
}

func TestListServersParams_SetOffset(t *testing.T) {
	t.Parallel()

	type args struct {
		offset int
	}

	tests := []struct {
		name        string
		args        args
		expectedErr error
	}{
		{
			name: "OK",
			args: args{
				offset: 100,
			},
		},
		{
			name: "ERR: offset < 0",
			args: args{
				offset: -1,
			},
			expectedErr: domain.ValidationError{
				Field: "offset",
				Err: domain.MinGreaterEqualError{
					Min:     0,
					Current: -1,
				},
			},
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			params := domain.NewListServersParams()

			require.ErrorIs(t, params.SetOffset(tt.args.offset), tt.expectedErr)
			if tt.expectedErr != nil {
				return
			}
			assert.Equal(t, tt.args.offset, params.Offset())
		})
	}
}

type serverKeyValidationTest struct {
	name        string
	key         string
	expectedErr error
}

func newServerKeyValidationTests() []serverKeyValidationTest {
	return []serverKeyValidationTest{
		{
			name: "ERR: server key length < 1",
			expectedErr: domain.LenOutOfRangeError{
				Min: 1,
				Max: 10,
			},
		},
		{
			name: "ERR: server key length > 10",
			key:  "keykeykeyke",
			expectedErr: domain.LenOutOfRangeError{
				Min:     1,
				Max:     10,
				Current: len("keykeykeyke"),
			},
		},
	}
}
